/// ReWindowLibrary
/// A C++20 Library for creating and managing windows in a platform-independent manner
/// Developed and Maintained by the boys @ Redacted Software.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file gamepadbutton.h
/// @desc GamepadButton class and enumerations to define standard buttons found on a Gamepad
/// @edit 2024-07-29

#include <string>
#include <utility>
#include <J3ML/LinearAlgebra.hpp>
namespace ReWindow {
    class GamepadButton;
    class GamepadTrigger;
    class GamepadThumbstick;
}

class ReWindow::GamepadButton {
protected:
    std::string mnemonic_btn_code;
public:
    explicit GamepadButton(std::string mnemonic) : mnemonic_btn_code(std::move(mnemonic)){}
    [[nodiscard]] std::string GetMnemonicButtonCode() const;

    /// Compares two GamepadButtons by their mnemonic button codes, not their activation state.
    bool operator ==(const GamepadButton& rhs) const;
};

class ReWindow::GamepadTrigger {
protected:
    std::string mnemonic_trigger_code;
    /// A float value between 0 and 1 representing how much the trigger has been pushed in by.
    float position = 0.0f;

    /// The minimum possible movement of the trigger from the at-rest position.
    /// Movements less than this won't count.
    float dead_zone = 0.0f;
public:
    void SetDeadzone(float minimum = 0.01f);
    [[nodiscard]] float GetActuation() const;
public:
    explicit GamepadTrigger(std::string mnemonic) : mnemonic_trigger_code(std::move(mnemonic)){}
};

class ReWindow::GamepadThumbstick {
protected:
    std::string mnemonic_stick_code;
    // X -1 is all the way left, X +1 is all the way right.
    // Y -1 is all the way down, Y +1 is all the way up.
    Vector2 position = {0.0f, 0.0f};

    // The minimum possible movement from the center in any direction.
    float dead_zone = 0.0f;
public:
    [[nodiscard]] Vector2 GetPosition() const;
    void SetDeadzone(float minimum = 0.01f);
public:
    explicit GamepadThumbstick(std::string mnemonic) : mnemonic_stick_code(std::move(mnemonic)){}
};



using J3ML::LinearAlgebra::Vector2;

namespace ReWindow::GamepadButtons {
    static const GamepadButton Triangle("△");
    static const GamepadButton Square("□");
    static const GamepadButton Circle("○");
    static const GamepadButton Cross("X");

    // If you like xbox :shrug:
    static const GamepadButton Y = Triangle;
    static const GamepadButton X = Square;
    static const GamepadButton A = Cross;
    static const GamepadButton B = Circle;

    static const GamepadButton LButton("LB");
    static const GamepadButton RButton("RB");

    /* For controllers where L2 & R2 is a button or counts as one when pressed all the way down.
     * Gamecube & PS2 controllers do this. - Redacted. */
    static const GamepadButton LButton2("LB2");
    static const GamepadButton RButton2("RB2");

    // The buttons when you press in the sticks.
    static const GamepadButton LButton3("LB3");
    static const GamepadButton RButton3("RB3");

    // Will eventually need to handle making it not possible to, for ex. Press left and right at the same time.
    static const GamepadButton DPadUp("DPU");
    static const GamepadButton DPadDown("DPD");
    static const GamepadButton DPadLeft("DPL");
    static const GamepadButton DPadRight("DPR");

    static const GamepadButton Select("SEL");
    static const GamepadButton Start("START");
}

namespace ReWindow::GamepadTriggers {
    static const ReWindow::GamepadTrigger Left {"LT"};
    static const ReWindow::GamepadTrigger Right {"RT"};
}

namespace ReWindow::GamepadThumbsticks {
    static const GamepadThumbstick Left {"LS"};
    static const GamepadThumbstick Right {"RS"};
}