#pragma once
#include <cstdint>
#include <vector>
#include <Event.h>
#include <map>
#include <thread>
#include <rewindow/types/key.h>
#include <rewindow/types/cursors.h>
#include <rewindow/types/mousebutton.h>
#include <rewindow/types/gamepadbutton.h>
#include <J3ML/LinearAlgebra.hpp>
#include <rewindow/types/WindowEvents.hpp>
#include <format>
#include <queue>

#include <experimental/propagate_const>

// Figure out what to do with this shit
enum class RWindowFlags: uint8_t {
    IN_FOCUS,
    FULLSCREEN,
    RESIZABLE,
    VSYNC,
    QUIT,
    MAX_FLAG
};

// Just throwing this in the header for now
std::string RWindowFlagToStr(RWindowFlags flag); /*{
        switch (flag) {
            case RWindowFlags::IN_FOCUS: return "IN_FOCUS";
            case RWindowFlags::FULLSCREEN: return "FULLSCREEN";
            case RWindowFlags::RESIZABLE: return "RESIZEABLE";
            case RWindowFlags::VSYNC: return "VSYNC";
            case RWindowFlags::QUIT: return "QUIT";
            case RWindowFlags::MAX_FLAG: return "MAX_FLAG";
            default:
                return "unimplemented flag";
        }
    };
    */

using J3ML::LinearAlgebra::Vector2;

namespace ReWindow {


    class KeyboardState {
    public:
        std::map<Key, bool> PressedKeys;
    };

    class GamepadState {
    public:
        std::map<GamepadButton, bool> PressedButtons;
    };

    class MouseState {
    public:
        struct
        {
            bool LMB = false;
            bool RMB = false;
            bool MMB = false;
            bool SideButton1 = false;
            bool SideButton2 = false;
            bool MWheelUp = false;
            bool MWheelDown = false;
        } Buttons;


        Vector2 Position;
        int Wheel = 0;

        [[nodiscard]] bool IsDown(const MouseButton& btn) const
        {
            if (btn == MouseButtons::Left) return Buttons.LMB;
            if (btn == MouseButtons::Right) return Buttons.RMB;
            if (btn == MouseButtons::Middle) return Buttons.MMB;
            if (btn == MouseButtons::Mouse4) return Buttons.SideButton1;
            if (btn == MouseButtons::Mouse5) return Buttons.SideButton2;
            //if (btn == MouseButtons::MWheelUp) return Buttons.MWheelUp;
            //if (btn == MouseButtons::MWheelDown) return Buttons.MWheelDown;

            return false; // Unknown button?
        }

        void Set(const MouseButton& btn, bool state)
        {
            if (btn == MouseButtons::Left) Buttons.LMB = state;
            if (btn == MouseButtons::Right) Buttons.RMB = state;
            if (btn == MouseButtons::Middle) Buttons.MMB = state;
            if (btn == MouseButtons::Mouse4) Buttons.SideButton1 = state;
            if (btn == MouseButtons::Mouse5) Buttons.SideButton2 = state;
            //if (btn == MouseButtons::MWheelUp) Buttons.MWheelUp = state;
            //if (btn == MouseButtons::MWheelDown) Buttons.MWheelDown = state;
        }

        bool& operator[](const MouseButton& btn)
        {
            if (btn == MouseButtons::Left) return Buttons.LMB;
            if (btn == MouseButtons::Right) return Buttons.RMB;
            if (btn == MouseButtons::Middle) return Buttons.MMB;
            if (btn == MouseButtons::Mouse4) return Buttons.SideButton1;
            if (btn == MouseButtons::Mouse5) return Buttons.SideButton2;
            //if (btn == MouseButtons::MWheelUp) return Buttons.MWheelUp;
            //if (btn == MouseButtons::MWheelDown) return Buttons.MWheelDown;

            throw std::runtime_error("Attempted to handle unmapped mouse button");
        }
    };

    enum class RenderingAPI: uint8_t {
        OPENGL = 0,
        //Vulkan is unimplemented.
        VULKAN = 1,
    };
    //

    class RWindowImpl;
    class RWindow;
}

class ReWindow::RWindowImpl {
private:
    // Class for platform specific "global" variables, information, functionality, etc
    // It is private as it should not be accessed outside the Impl. It has also
    // purposefully been left undefined, so implementors can define it as they see fit.
    // Make sure to do proper cleanup of the class and pointer if utilized. It should
    // be handled with care.
    //
    // Example for initialization:
    // RWindowImpl::RWindowImpl(ARGS) : pPlatform( new Platform ) { XYZ }
    //
    // Example for destruction:
    // RWindowImpl::~RWindowImpl() {
    //    if (pPlatform) { delete pPlatform; };
    // }
    //
    // Example for definition:
    // class RWindowImpl::Platform {
    // public:
    //    Platform() = default;
    // public:
    //    int a;
    //    int b;
    //    int c;
    // };
    //
    // - Maxine
    class Platform;
    Platform* pPlatform = nullptr;
protected: // Should maybe make private
    int width = 1280;
    int height = 720;

    bool open = false; // Is the underlying OS-Window-Handle actually open.
    bool resizable = true;
    bool fullscreen_mode = false;
    bool focused = true;
    bool vsync = false;
    bool cursor_visible = true;
    bool closing = false;

    float delta_time = 0.f;
    float refresh_rate = 0.f;
    unsigned int refresh_count = 0;

    std::string title = "Redacted Window";

    Vector2 lastKnownWindowSize {0, 0};

    // TODO: Implement ringbuffer / circular vector class of some sort.
    float refresh_rate_prev_1 = 0.f;
    float refresh_rate_prev_2 = 0.f;
    float refresh_rate_prev_3 = 0.f;
    float refresh_rate_prev_4 = 0.f;
    float refresh_rate_prev_5 = 0.f;

    float avg_refresh_rate = 0.0f;

    // Figure out what to do with this shit later
    bool flags[5];
    KeyboardState currentKeyboard;  // current frame keyboard state.
    KeyboardState previousKeyboard; // previous frame keyboard state.
    MouseState currentMouse;		// purrent frame mouse state.
    MouseState previousMouse;		// previous frame mouse state
    RenderingAPI renderer = RenderingAPI::OPENGL;
    //
public:
    explicit RWindowImpl(const std::string& wTitle,
                         int wWidth = 640, int wHeight = 480,
                         RenderingAPI wRenderer = RenderingAPI::OPENGL,
                         bool wFullscreen = false,
                         bool wResizable = true,
                         bool wVsync = false);
    ~RWindowImpl();
public:
    void Open();
    void Close();
    void PollEvents();
    void Refresh();
public:
    bool IsFocused() const;
    bool IsFullscreen() const;
    bool IsResizable() const;
    bool IsVsyncEnabled() const;
    bool IsAlive() const;
public:
    void SetFullscreen(bool fs);
    void SetResizable(bool fs);
    void SetVsyncEnabled(bool vsync);
    void SetSize(int width, int height);
    void SetSize(const Vector2& size);
    // Added here for now
    void SetPos(int x, int y);
    void SetPos(const Vector2& pos);
    //
    void SetTitle(const std::string& title);
public:
    std::string GetTitle() const;
    Vector2 GetPos() const;
    Vector2 GetSize() const;
    int GetWidth() const;
    int GetHeight() const;
    float GetDeltaTime() const;
    float GetRefreshRate() const;
    float GetRefreshCounter() const;
public:
    // Figure out what to do with these later
    void Raise() const;
    void Lower() const;
    void DestroyOSWindowHandle();
    void SetCursorVisible(bool cursor_enable);
    Vector2 GetAccurateMouseCoordinates() const;
    void GLSwapBuffers();
    void Fullscreen();
    void RestoreFromFullscreen();
    // I know this doesn't modify the class, but it indirectly modifies the window
    // Making it const just seems deceptive.
    void SetCursorStyle(CursorStyle style) const;
    Vector2 GetPositionOfRenderableArea() const;
    std::string getGraphicsDriverVendor();
    void SetFlag(RWindowFlags flag, bool state);
    void processKeyRelease (Key key);
    void processKeyPress (Key key);
    void processOnClose();
    void processOnOpen();
    void processMousePress(const MouseButton& btn);
    void processMouseRelease(const MouseButton& btn);
    void processFocusIn();
    void processFocusOut();
    void processMouseMove(Vector2 last_pos, Vector2 new_pos);
    void processMouseWheel(int scrolls);
    virtual void OnOpen() {}
    virtual void OnClosing() {}
    virtual void OnFocusLost(const RWindowEvent& e) {}
    virtual void OnFocusGain(const RWindowEvent& e) {}
    virtual void OnRefresh(float elapsed) {}
    virtual void OnResizeSuccess() {}
    virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}
    virtual void OnKeyDown(const KeyDownEvent&) {}
    virtual void OnKeyUp(const KeyUpEvent&) {}
    virtual void OnMouseMove(const MouseMoveEvent&) {}
    virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}
    virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}
    virtual void OnMouseWheel(const MouseWheelEvent&) {}
    Event<> OnOpenEvent;
    Event<> OnClosingEvent;
    Event<RWindowEvent> OnFocusLostEvent;
    Event<RWindowEvent> OnFocusGainEvent;
    Event<float> OnRefreshEvent;
    Event<WindowResizeRequestEvent> OnResizeRequestEvent;
    Event<KeyDownEvent> OnKeyDownEvent;
    Event<KeyUpEvent> OnKeyUpEvent;
    Event<MouseMoveEvent> OnMouseMoveEvent;
    Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
    Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
    Event<MouseWheelEvent> OnMouseWheelEvent;

    Vector2 GetMouseCoordinates() const;
    bool GetFlag(RWindowFlags flag) const;
    //bool IsAlive() const;
    void SetSizeWithoutEvent(const Vector2& size);
    std::vector<RWindowEvent> eventLog;
    std::queue<RWindowEvent> eventQueue;
    void LogEvent(const RWindowEvent& e) { eventLog.push_back(e);};
    RWindowEvent GetLastEvent() const {
        return eventLog.back();
    };
    void SetLastKnownWindowSize(const Vector2& size);
    void SetRenderer(RenderingAPI api);
    bool IsKeyDown(Key key) const;
    bool IsMouseButtonDown(const MouseButton &button) const;
    void ManagedRefresh();
    float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end);
    std::chrono::steady_clock::time_point GetTimestamp();
    void UpdateFrameTiming(float frame_time);
    /*
    bool IsResizable() const;
    bool IsFullscreen() const;
    bool IsFocused() const;
    bool IsVsyncEnabled() const;
     */
    void ForceClose();
    void ForceCloseAndTerminateProgram();
    int GetMouseWheelPersistent() const { return currentMouse.Wheel;}
    [[nodiscard]] bool IsOpen() const { return open;}
    [[nodiscard]] bool IsClosing() const { return closing;}


};

class ReWindow::RWindow : private RWindowImpl {
public:
    //RWindow() = default;
    //RWindow();
    explicit RWindow(const std::string& wTitle,
                     int wWidth = 640, int wHeight = 480,
                     RenderingAPI wRenderer = RenderingAPI::OPENGL,
                     bool wFullscreen = false,
                     bool wResizable = true,
                     bool wVsync = false) :
                     RWindowImpl(wTitle, wWidth, wHeight, wRenderer, wFullscreen, wResizable, wVsync) {};
    ~RWindow() = default;
public:
    // Platform dependant
    void Open() { RWindowImpl::Open(); };
    void Close() { RWindowImpl::Close(); };
    void PollEvents() { RWindowImpl::PollEvents(); };
    void Refresh() { RWindowImpl::Refresh(); };
public:
    // Shared
    /// Returns whether the window currently has mouse and/or keyboard focus.
    [[nodiscard]] bool IsFocused() const { return RWindowImpl::IsFocused(); };
    /// Returns whether the window is currently in Fullscreen.
    // TODO: Support Fullscreen, FullscreenWindowed, and Windowed?
    [[nodiscard]] bool IsFullscreen() const { return RWindowImpl::IsFullscreen(); } ;
    /// Returns whether the window can be resized.
    [[nodiscard]] bool IsResizable() const { return RWindowImpl::IsResizable(); } ;
    /// Returns whether V-Sync is enabled.
    [[nodiscard]] bool IsVsyncEnabled() const { return RWindowImpl::IsVsyncEnabled(); } ;
    /// Returns whether the window is considered to be alive. Once dead, any logic loop should be terminated, and the cleanup procedure should run.
    [[nodiscard]] bool IsAlive() const { return RWindowImpl::IsAlive(); } ;
    // Should have IsOpen since window could be closed then reopened
public:
    // Platform dependant
    /// Sets whether or not to make the window fullscreen.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetFullscreen(bool fs) { RWindowImpl::SetFullscreen(fs); };
    /// Sets whether or not to make the window resizable.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetResizable(bool resizable) { RWindowImpl::SetResizable(resizable); };
    /// Sets whether or not to enable vertical synchronization.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetVsyncEnabled(bool vsync) { RWindowImpl::SetVsyncEnabled(vsync); };
    /// Requests the operating system to change the window size.
    /// @param width
    /// @param height
    void SetSize(int width, int height) { RWindowImpl::SetSize(width, height); };
    /// Requests the operating system to change the window size.
    /// @param size
    void SetSize(const Vector2& size) { RWindowImpl::SetSize(size); };

    // Shared
    /// Sets the title of this window.
    void SetTitle(const std::string& title) { RWindowImpl::SetTitle(title); };
public:
    // Shared
    [[nodiscard]] std::string GetTitle() const { return RWindowImpl::GetTitle(); } ;
    /// Returns the position of the window's top-left corner relative to the display
    [[nodiscard]] Vector2 GetPos() const { return RWindowImpl::GetPos(); } ;
    /// Returns the known size of the window, in {x,y} pixel measurement.
    [[nodiscard]] Vector2 GetSize() const { return RWindowImpl::GetSize(); } ;
    /// Returns the horizontal length of the renderable area in pixels.
    [[nodiscard]] int GetWidth() const { return RWindowImpl::GetWidth(); } ;
    /// Returns the vertical length of the renderable area, in pixels.
    [[nodiscard]] int GetHeight() const { return RWindowImpl::GetHeight(); } ;
    /// Returns the amount of time, in seconds, between the current and last frame.
    /// Technically, no, it returns the elapsed time of the frame, start to finish.
    [[nodiscard]] float GetDeltaTime() const { return RWindowImpl::GetDeltaTime(); } ;
    /// Returns the approximate frames-per-second using delta time.
    [[nodiscard]] float GetRefreshRate() const { return RWindowImpl::GetRefreshRate(); } ;
    /// Returns the number of frames ran since the windows' creation.
    [[nodiscard]] float GetRefreshCounter() const { return RWindowImpl::GetRefreshCounter(); } ;
public:
    // Figure out what to do with these later
    void Raise() const { RWindowImpl::Raise(); };
    void Lower() const { RWindowImpl::Lower(); };
    void DestroyOSWindowHandle() { RWindowImpl::DestroyOSWindowHandle(); };
    void SetCursorVisible(bool cursor_enable) { RWindowImpl::SetCursorVisible(cursor_enable); };
    Vector2 GetAccurateMouseCoordinates() const { return RWindowImpl::GetAccurateMouseCoordinates(); } ;
    void GLSwapBuffers() { RWindowImpl::GLSwapBuffers(); };
    void Fullscreen() { RWindowImpl::Fullscreen(); };
    void RestoreFromFullscreen() { RWindowImpl::RestoreFromFullscreen(); };
    // I know this doesn't modify the class, but it indirectly modifies the window
    // Making it const just seems deceptive.
    void SetCursorStyle(CursorStyle style) const { return RWindowImpl::SetCursorStyle(style); } ;
    Vector2 GetPositionOfRenderableArea() const { return RWindowImpl::GetPositionOfRenderableArea(); } ;
    std::string getGraphicsDriverVendor() { return RWindowImpl::getGraphicsDriverVendor(); };
    void SetFlag(RWindowFlags flag, bool state) { RWindowImpl::SetFlag(flag, state); };
    void processKeyRelease (Key key) { RWindowImpl::processKeyRelease(key); };
    void processKeyPress (Key key) { RWindowImpl::processKeyPress(key); };
    void processOnClose() { RWindowImpl::processOnClose(); };
    void processOnOpen() { RWindowImpl::processOnOpen(); };
    void processMousePress(const MouseButton& btn) { RWindowImpl::processMousePress(btn); };
    void processMouseRelease(const MouseButton& btn) { RWindowImpl::processMouseRelease(btn); };
    void processFocusIn() { RWindowImpl::processFocusIn(); };
    void processFocusOut() { RWindowImpl::processFocusOut(); };
    void processMouseMove(Vector2 last_pos, Vector2 new_pos) { RWindowImpl::processMouseMove(last_pos, new_pos); };
    void processMouseWheel(int scrolls) { RWindowImpl::processMouseWheel(scrolls); };
    virtual void OnOpen() {}
    virtual void OnClosing() {}
    virtual void OnFocusLost(const RWindowEvent& e) {}
    virtual void OnFocusGain(const RWindowEvent& e) {}
    virtual void OnRefresh(float elapsed) {}
    virtual void OnResizeSuccess() {}
    virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}
    virtual void OnKeyDown(const KeyDownEvent&) {}
    virtual void OnKeyUp(const KeyUpEvent&) {}
    virtual void OnMouseMove(const MouseMoveEvent&) {}
    virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}
    virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}
    virtual void OnMouseWheel(const MouseWheelEvent&) {}
    Event<> OnOpenEvent;
    Event<> OnClosingEvent;
    Event<RWindowEvent> OnFocusLostEvent;
    Event<RWindowEvent> OnFocusGainEvent;
    Event<float> OnRefreshEvent;
    Event<WindowResizeRequestEvent> OnResizeRequestEvent;
    Event<KeyDownEvent> OnKeyDownEvent;
    Event<KeyUpEvent> OnKeyUpEvent;
    Event<MouseMoveEvent> OnMouseMoveEvent;
    Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
    Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
    Event<MouseWheelEvent> OnMouseWheelEvent;

    Vector2 GetMouseCoordinates() const { return RWindowImpl::GetMouseCoordinates(); };
    bool GetFlag(RWindowFlags flag) const { return RWindowImpl::GetFlag(flag); };
    //bool IsAlive() const { return RWindowImpl::IsAlive(); };
    void SetSizeWithoutEvent(const Vector2& size) { RWindowImpl::SetSizeWithoutEvent(size); };
    void LogEvent(const RWindowEvent& e) { RWindowImpl::LogEvent(e); };
    RWindowEvent GetLastEvent() const { return RWindowImpl::GetLastEvent(); }
    void SetLastKnownWindowSize(const Vector2& size) { RWindowImpl::SetLastKnownWindowSize(size); };
    void SetRenderer(RenderingAPI api) { RWindowImpl::SetRenderer(api); };
    bool IsKeyDown(Key key) const { return RWindowImpl::IsKeyDown(key); };
    bool IsMouseButtonDown(const MouseButton &button) const { return RWindowImpl::IsMouseButtonDown(button); };
    void ManagedRefresh() { RWindowImpl::ManagedRefresh(); };
    float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end) { return RWindowImpl::ComputeElapsedFrameTimeSeconds(start, end); };
    std::chrono::steady_clock::time_point GetTimestamp() { return RWindowImpl::GetTimestamp(); };
    void UpdateFrameTiming(float frame_time) { RWindowImpl::UpdateFrameTiming(frame_time); };
    //bool IsResizable() const { return RWindowImpl::IsResizable(); };
    //bool IsFullscreen() const { return RWindowImpl::IsFullscreen(); };
    //bool IsFocused() const { return RWindowImpl::IsFocused(); };
    //bool IsVsyncEnabled() const { return RWindowImpl::IsVsyncEnabled(); };
    void ForceClose() { RWindowImpl::ForceClose(); };
    void ForceCloseAndTerminateProgram() { RWindowImpl::ForceCloseAndTerminateProgram(); };
    int GetMouseWheelPersistent() const { return RWindowImpl::GetMouseWheelPersistent();}
    [[nodiscard]] bool IsOpen() const { return RWindowImpl::IsOpen();}
    [[nodiscard]] bool IsClosing() const { return RWindowImpl::IsClosing();}

};